import pytest
from a1 import find_minimum
import time
import numpy as np
from scipy.stats import multivariate_normal as mvn


def min_brute_force(matrix):
    """
    Generic brute force implementation of an algorithm to
    find the minimum in a 2D matrix.

    Args:
        matrix (np.array): 2D matrix.

    Returns:
        (int, int): tuple of indices (i, j).
    """
    min_idx = 0, 0
    min = matrix[min_idx]
    for i in range(len(matrix)):
        for j in range(len(matrix[0])):
            if matrix[i, j] < min:
                min_idx = i, j
                min = matrix[i, j]
    return min_idx


def get_data(start=-2, stop=2, step=0.001):
    """Get a grid of values from negative bivariate normal distribution.

    We first create (x,y) coordinates for each point in the grid,
    and get values from the bivariate normal PDF.

    Gaussian PDF is not special here. We could use any concave
    function, e.g., y = x^2 + y^2.
    """
    f = mvn(mean=(0,0), cov=1).pdf
    xy = np.dstack(np.mgrid[start:stop:step, start:stop:step])
    return -f(xy)

def test_correctness():
    """Test the correctness of the algorithm. """
    M = get_data(start=-2, stop=2, step=0.1)
    # min at center 
    assert find_minimum(M) == min_brute_force(M)
    # min at the edges
    for m in (M[:M.shape[0]//2, :], M[M.shape[0]//2:, :], 
              M[:, :M.shape[0]//2], M[:, M.shape[0]//2:]): 
        assert find_minimum(m) == min_brute_force(m)
    # min at the corners
    for m in (M[M.shape[0]//2:, M.shape[1]//2:],
            M[:M.shape[0]//2, M.shape[1]//2:],
            M[M.shape[0]//2:, :M.shape[1]//2],
            M[:M.shape[0]//2, :M.shape[1]//2]):
        assert find_minimum(m) == min_brute_force(m)

def test_performance():
    """Test the performance: it should work faster than generic min().
        
        It is unlikely, but since the timing depends on other processes
        running on the system, there may be false positives here.
    """
    M = get_data()
    tr = list(zip(np.random.randint(0, M.shape[0]//2, size=5),
                np.random.randint(0, M.shape[0]//2, size=5)))
    start = time.time()
    for i, j in tr:
        _ = min_brute_force(M[i:, j:])
    brute_force_time = time.time() - start

    start = time.time()
    for i, j in tr:
        _ = find_minimum(M[i:, j:])
    our_time = time.time() - start

    assert our_time < brute_force_time
